@@ -1,5 +1,5 @@
 /*
- * $Id: download.c,v 1.272 2004/01/19 18:42:55 hipnod Exp $
+ * $Id: download.c,v 1.278 2004/11/04 23:17:51 mkern Exp $
  *
  * Copyright (C) 2001-2003 giFT project (gift.sourceforge.net)
  *
@@ -141,7 +141,7 @@ size_t download_throttle (Chunk *chunk, size_t len)
 	/* get the number of bytes remaining for this chunk so that we can
 	 * clamp len properly */
 	rem = chunk->stop - (chunk->start + chunk->transmit);
-	len = (size_t)(MIN (rem, len));
+	len = (size_t)(MIN (rem, (off_t) len));
 
 #ifdef THROTTLE_ENABLE
 	return chunk_throttle (chunk, MAX_DOWNLOAD_BW, len, download_credits);
@@ -224,15 +224,119 @@ static char *uniq_file (Transfer *transfer)
 	return uniq;
 }
 
+#ifdef _MSC_VER
+
+/* Max additional length of unique incoming prefix. */
+#define MAX_UNIQ_LEN 21 /* "%04lX%08lX%08lX.%s" */
+
+/* Max additional length if completed filename already exists */
+#define MAX_UNIQ_FILENAME_LEN 3 /* strlen (".xx") */
+
+/* Max additional length of state file postfix. */
+#define MAX_STATE_LEN 6 /* strlen (".state") */
+
+/* Max additional length of corrupted dir. */
+#define MAX_CORRUPTED_LEN 10 /* strlen ("corrupted/") */
+
+/* Minimum number of chars which must remain in filename. */
+#define MIN_FILENAME_REMAIN 5
+
+/*
+ * Makes a best effort to return a copy of save filename which complies
+ * with various path length limits in Microsoft's runtime.
+ */
+static char *msvc_limit_filename (const char *filename)
+{
+	int max_len = 0, len;
+	char *ext, *file;
+
+	/* Make a copy to modify. */
+	file = strdup (filename);
+
+	/*
+	 * With MSVC the maximum path length must not exceed _MAX_PATH or fopen
+	 * will fail. We must thus limit the filename so neither of the paths
+	 * we are going to use hits this limit. We do this by first finding the
+	 * maximum path length for the given filename and then removing the
+	 * difference to the allowed path length from the back of filename while
+	 * preserving the extension.
+	 */
+
+	/* 
+	 * In the incoming dir the state path for corrupted files will be
+	 * longest.
+	 */
+	max_len = strlen (INCOMING_PATH (("%s", file)));
+	max_len += MAX_UNIQ_LEN + MAX_STATE_LEN + MAX_CORRUPTED_LEN;
+
+	/* 
+	 * In the completed dir the final path with unique filename postfix will
+	 * be longest.
+	 */
+	len = strlen (COMPLETED_PATH (("%s", file)));
+	len += MAX_UNIQ_FILENAME_LEN;
+
+	if (len > max_len)
+		max_len = len;
+
+	/* If we are withing the limit there is nothing to do. */
+	len = max_len - _MAX_PATH;
+
+	if (len <= 0)
+		return file;
+
+	/*
+	 * If filename would be shorter than acceptable minimum after shortening
+	 * don't modify it and let fopen fail later.
+	 */
+	if (len + MIN_FILENAME_REMAIN >= (int) strlen (file))
+	{
+		GIFT_WARN (("Cannot shorten '%s' to meet _MAX_PATH requirements.",
+		           file));
+		return file;
+	}
+
+	/* Try to preserve extension. */
+	if (!(ext = strrchr (file, '.')) || (int) strlen (ext) >= len)
+	{
+		ext = file + strlen (file);
+	}
+
+	/* And kill ending of file */
+	memmove (file + strlen (file) - len - strlen (ext), ext, strlen (ext) + 1);
+
+	GIFT_TRACE (("Shortened '%s' to '%s' to meet _MAX_PATH requirements.",
+	            filename, file));
+
+	return file;
+}
+#endif
+
 Transfer *download_new (TCPC *c, if_event_id requested,
                         char *uniq, char *filename, char *hash,
                         off_t size)
 {
 	Transfer *transfer;
 
+#ifdef _MSC_VER
+	/*
+	 * If uniq is not NULL this is a resumed transfer with transfer->path
+	 * being loaded from the state file at a later point and we should not
+	 * mess with the filename.
+	 */
+	if (uniq)
+		filename = strdup (filename);
+	else
+		filename = msvc_limit_filename (filename);
+#endif
+
 	if (!(transfer = transfer_new (TRANSFER_DOWNLOAD, filename, hash, size)))
 		return NULL;
 
+#ifdef _MSC_VER
+	free (filename);
+#endif
+
 	/* write immediately */
 	transfer->flag = STATE_INTERVAL;
 
@@ -432,14 +536,27 @@ void download_free (Transfer *transfer, int premature)
 
 static char *calculate_final_path (Transfer *transfer)
 {
-	char        *final;
+	char        *final, *ext;
 	char         fmt[PATH_MAX];
+	char         filename[PATH_MAX];
 	unsigned int uniq = 0;
 
+	/*
+	 * FIXME: Should use strlcpy here, or better yet, invent some sort of
+	 * strcspn_rev() and use it to determine the length of the name without
+	 * the extension, thus eliminating the need for a copied object.
+	 */
+	assert (strlen (transfer->filename) < sizeof (filename));
+	strcpy (filename, transfer->filename);
+
+	if ((ext = strrchr (filename, '.')))
+		*ext++ = 0;
+
 	do
 	{
-		snprintf (fmt, sizeof (fmt) - 1, "%s%s%s", transfer->filename,
-		          (uniq ? "." : ""), (uniq ? ITOA (uniq) : ""));
+		snprintf (fmt, sizeof (fmt) - 1, "%s%s%s%s%s", filename,
+		          (uniq ? "." : ""), (uniq ? ITOA (uniq) : ""),
+		          (ext ? "." : ""), (ext ? ext : ""));
 
 		final = COMPLETED_PATH (("%s", fmt));
 		uniq++;
@@ -626,6 +743,7 @@ static BOOL source_pause (Source *source, Transfer *transfer)
 	return TRUE;
 }
 
+/* this handles pause requests by the FE _only_ */
 void download_pause (Transfer *transfer)
 {
 	DOWNLOAD(transfer)->paused = TRUE;
@@ -633,17 +751,36 @@ void download_pause (Transfer *transfer)
 	list_foreach (transfer->chunks,  (ListForeachFunc) chunk_pause, transfer);
 	list_foreach (transfer->sources, (ListForeachFunc) source_pause, transfer);
 
+	/* save state file so the paused state gets written to it */
+	if (!(download_state_save (transfer)))
+	{
+		handle_write_error (transfer);
+		return;
+	}
+
 	/*
 	 * Drop all in-memory state, including the chunks, as the state in them
-	 * may not reflect the actual state on disk.
+	 * may not reflect the actual state on disk. If a download is paused it
+	 * has no chunks and the file descriptor is closed.
 	 */
 	download_state_rollback (transfer);
 }
 
 static void handle_write_error (Transfer *transfer)
 {
-	/* revert the download to the last good state */
-	download_pause (transfer);
+	/* pause download and revert to the last good state */
+	DOWNLOAD(transfer)->paused = TRUE;
+
+	list_foreach (transfer->chunks,  (ListForeachFunc) chunk_pause, transfer);
+	list_foreach (transfer->sources, (ListForeachFunc) source_pause, transfer);
+
+	/*
+	 * Drop all in-memory state, including the chunks, as the state in them
+	 * may not reflect the actual state on disk. Note that we have no way to
+	 * write the paused state to disk so the download will be resumed on next
+	 * restart (failing again if the disk is still full).
+	 */
+	download_state_rollback (transfer);
 
 	/* update front ends because source_status_set won't */
 	if_transfer_change (transfer->event, TRUE);
@@ -673,6 +810,17 @@ void download_unpause (Transfer *transfer)
 	/* restore the chunks */
 	download_state_initialize (transfer);
 
+	/*
+	 * Save state file again so the new paused state gets written to it. If
+	 * this fails the user still has no free disk space and we go back to
+	 * paused again.
+	 */
+	if (!(download_state_save (transfer)))
+	{
+		handle_write_error (transfer);
+		return;
+	}
+
 	list_foreach (transfer->sources, (ListForeachFunc) source_unpause, transfer);
 }
 
@@ -1230,6 +1378,11 @@ static int add_source (Transfer *transfer, Source *source)
 	if (DOWNLOAD(transfer)->paused)
 	{
 		source_status_set (source, SOURCE_PAUSED, NULL);
+
+		/* Tell the front end about the new source. source_status_set will not
+		 * send anything to the FE because source->chunk is NULL.
+		 */
+		if_transfer_addsource (transfer->event, source);
 		return TRUE;
 	}
 
@@ -1287,15 +1440,15 @@ BOOL download_make_source (Transfer *transfer,
                            char *user, char *hash, char *url)
 {
 	List     *ptr;
-	Source   *tmp;
+	Source   *new_src;
 	Protocol *p;
 
-	if (!(tmp = source_new (user, hash, transfer->total, url)))
+	if (!(new_src = source_new (user, hash, transfer->total, url)))
 		return FALSE;
 
-	if (!(p = tmp->p))
+	if (!(p = new_src->p))
 	{
-		source_free (tmp);
+		source_free (new_src);
 		return FALSE;
 	}
 
@@ -1307,17 +1460,18 @@ BOOL download_make_source (Transfer *transfer,
 			continue;
 
 		/* forget it if the source is already there */
-		if (p->source_cmp (p, tmp, source) == 0)
+		if (p->source_cmp (p, new_src, source) == 0)
 		{
-			source_free (tmp);
+			source_free (new_src);
 			return FALSE;
 		}
 	}
 
-	source_free (tmp);
-
-	/* this may fail, but doesn't affect return value */
-	download_add_source (transfer, user, hash, url);
+	if (add_source (transfer, new_src) == FALSE)
+	{
+		source_free (new_src);
+		return FALSE;
+	}
 
 	return TRUE;
 }
