@@ -1,7 +1,7 @@
 /*
- * $Id: giftd.c,v 1.24 2003/10/16 18:50:55 jasta Exp $
+ * $Id: giftd.c,v 1.30 2004/11/12 00:58:21 hexwab Exp $
  *
- * Copyright (C) 2001-2003 giFT project (gift.sourceforge.net)
+ * Copyright (C) 2001-2004 giFT project (gift.sourceforge.net)
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -101,7 +101,7 @@ static const char *msg_usage =
 "\n";
 
 static const char *msg_copyleft =
-"Copyright (c) 2001-2003 giFT project (http://giftproject.org/)\n"
+"Copyright (c) 2001-2004 giFT project (http://giftproject.org/)\n"
 "This is free software; see the source for copying conditions.  There is NO\n"
 "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"
 "\n";
@@ -436,6 +436,25 @@ static Config *check_config (void)
 	return conf;
 }
 
+static void close_config (void)
+{
+	/*
+	 * Writing to the config file and even setting config variables in
+	 * memory causes all kinds of subtle side effects. Don't do it until we
+	 * come up with a better config system.
+	 */
+#if 0
+	/* Save config. The only thing which changes at this time is the FE
+	 * changeable shares_hidden.
+	 */
+	config_write (gift_conf);
+#endif
+
+	/* And free it */
+	config_free (gift_conf);
+	gift_conf = NULL;
+}
+
 static int gift_init (void)
 {
 	/*
@@ -524,7 +543,7 @@ static int gift_finish (void)
 	plugin_unloadall ();               /* cleanly destroy all plugins */
 	mime_cleanup ();                   /* cleanup mime data structures */
 	platform_cleanup ();               /* im getting sick of this */
-	config_free (gift_conf);           /* redundant documentation */
+	close_config ();                   /* redundant documentation */
 	log_cleanup ();                    /* so i think ill stop */
 
 #ifdef USE_LTDL
@@ -623,6 +642,9 @@ static Protocol *load_plugin (char *proto)
 	if (!(p = plugin_init (path, name)))
 		return NULL;
 
+	GIFT_INFO (("Plugin %s (%s) successfully loaded and initialized",
+	            p->name, p->version_str));
+
 	/*
 	 * WARNING: HACK because we know init_shares immediately follows
 	 * load_protocols and it only needs them initialized!  The design needs
@@ -685,6 +707,15 @@ static int init_shares (void)
 		exit (EXIT_SUCCESS);
 	}
 
+	/* hide shares if config file says so */
+#if 0
+	if (upload_status () < 0)
+		upload_disable ();
+#else
+	if (config_get_int (gift_conf, "sharing/shares_hidden=0"))
+		upload_disable ();
+#endif
+
 	/* automatic resync timer */
 	share_init_timer (RESYNC_INTERVAL);
 
@@ -753,8 +784,11 @@ static int gift_begin (void)
 	 * also provide a default web interface for modifying more complex
 	 * parameters via a standardized HTTP/HTML interface.
 	 */
-	if (!init_interface ())
-		return FALSE;
+	if (!cl_idxonly)
+	{
+		if (!init_interface ())
+			return FALSE;
+	}
 
 	/*
 	 * Load all dynamic protocols and call the startup routines.  At some
@@ -765,6 +799,19 @@ static int gift_begin (void)
 	if (!load_protocols ())
 		return FALSE;
 
+	/*
+	 * Start up previous downloads (from the state records).  Currently no
+	 * other initialization is required to use the transfer subsystem.  Do
+	 * this before starting share indexing thread because our directory
+	 * iteration seems not to like simultaneous use from multiple threads.
+	 * TODO: Needs further investigation.
+	 */
+	if (!cl_idxonly)
+	{
+		if (!init_transfer ())
+			return FALSE;
+	}
+
 	/*
 	 * Automatically update (and read) ~/.giFT/shares.  This is done to force
 	 * users to keep their share index up to date.
@@ -772,13 +819,6 @@ static int gift_begin (void)
 	if (!init_shares ())
 		return FALSE;
 
-	/*
-	 * Start up previous downloads (from the state records).  Currently no
-	 * other initialization is required to use the transfer subsystem.
-	 */
-	if (!init_transfer ())
-		return FALSE;
-
 	/*
 	 * Initialize the optional perl extension interface.  At present, this
 	 * interface is extremely crude and not really useful to anyone.  I am
