@@ -1,5 +1,5 @@
 /*
- * $Id: meta_tag.c,v 1.2 2003/06/27 10:06:59 hipnod Exp $
+ * $Id: meta_tag.c,v 1.31 2004/06/11 08:34:29 hipnod Exp $
  *
  * Copyright (C) 20003 giFT project (gift.sourceforge.net)
  *
@@ -23,8 +23,6 @@
 #include "meta.h"
 #include "meta_tag.h"
 
-#include <ctype.h>
-
 /*****************************************************************************/
 
 /*
@@ -33,8 +31,6 @@
  * TODO:
  *		- parse 16-bit Unicode strings and convert those that are
  *		  really just big ASCII strings (or is that ISO-8859-1?)
- *		- MusicMatch jukebox tagging support
- *		- Lyrics3 tagging support
  *		- compressed frames; compressed metaframes (CDM)
  *
  * If anyone wants to do these things, please feel free.
@@ -51,8 +47,11 @@
 /* maximum amount of memory used in reading frames */
 #define DEFAULT_MAX_FRAME_LEN        131072
 
+/* maximum sane possible size of a frame: 28-bits */
+#define MAX_FRAME_SIZE               0x0fffffff
+
 /* improve readability */
-#define TAG_FILE(tag)                (tag)->file
+#define ID3_DATA(tag)                (tag)->id3
 
 /*****************************************************************************/
 
@@ -66,37 +65,30 @@
 
 /*****************************************************************************/
 
-struct tag_file
+struct id3_data
 {
 	FILE              *f;
 	char              *name;
-	off_t              size;
 	Dataset           *keys;
 	struct id3v1      *id3v1;
 	struct id3v2      *id3v2;
-#if 0
-	struct musicmatch *musicmatch;         /* TODO */
-	struct lyrics3    *lyrics3;            /* TODO */
-#endif
 	Share             *share;
 };
 
 struct id3v1
 {
-	struct tag_file *file;
-	off_t            start;
+	struct id3_data *id3;
 };
 
 struct id3v2
 {
-	struct tag_file *file;
-	off_t            start;              /* start position of tag */
+	struct id3_data *id3;
 	size_t           max_len;            /* max buffer size */
 
 	size_t           data_len;
 	size_t           ext_hdr_len;        /* length of the ext hdr */
 	size_t           size;               /* data_len + hdrs */
-	size_t           data_offset;        /* offset in tag, from tag->v2_start */
+	size_t           data_offset;        /* offset in file */
 	size_t           fake_offset;        /* equal to data_offset except when
 	                                      * tag is unsynchronized */
 	BOOL             unsynched;          /* whole tag unsynch-encoded */
@@ -124,14 +116,13 @@ struct id3v2_frame
 	BOOL               has_data_len;  /* has data length indicator */
 };
 
-/* callback for handling each frame */
-typedef void (frame_parse_t) (struct id3v2 *tag, struct id3v2_frame *frame);
-
 typedef struct frame_handler
 {
-	char           *v2_str;              /* 3 char id from id3v2.2  */
-	char           *v3_str;              /* 4 char id from id3v2.3+ */
-	frame_parse_t  *parse;               /* parse function */
+	char     v2_str[4];              /* 3 char id from id3v2.2  */
+	char     v3_str[5];              /* 4 char id from id3v2.3+ */
+
+	/* callback for handling each frame */
+	void   (*parse) (struct id3v2 *tag, struct id3v2_frame *frame);
 } frame_handler_t;
 
 /*****************************************************************************/
@@ -139,21 +130,21 @@ typedef struct frame_handler
 static void read_v2_frames   (struct id3v2 *tag);
 static void read_v1          (struct id3v1 *tag);
 
-static void set_meta         (struct tag_file *file, char *id, char *key,
+static void set_meta         (struct id3_data *id3, char *id, char *key,
                               char *value);
 
 /*****************************************************************************/
 
-#define ID3V2_FRAME_HANDLER(name) \
+#define v2_frame_handler(name) \
 	static void name (struct id3v2 *tag, struct id3v2_frame *frame)
 
-ID3V2_FRAME_HANDLER (id3v2_parse_TLEN);
-ID3V2_FRAME_HANDLER (id3v2_parse_COMM);
-ID3V2_FRAME_HANDLER (id3v2_parse_TIT2);
-ID3V2_FRAME_HANDLER (id3v2_parse_TALB);
-ID3V2_FRAME_HANDLER (id3v2_parse_TRCK);
-ID3V2_FRAME_HANDLER (id3v2_parse_TPE1);
-ID3V2_FRAME_HANDLER (id3v2_parse_TYER);
+v2_frame_handler (id3v2_parse_TLEN);
+v2_frame_handler (id3v2_parse_COMM);
+v2_frame_handler (id3v2_parse_TIT2);
+v2_frame_handler (id3v2_parse_TALB);
+v2_frame_handler (id3v2_parse_TRCK);
+v2_frame_handler (id3v2_parse_TPE1);
+v2_frame_handler (id3v2_parse_TYER);
 
 /*****************************************************************************/
 
@@ -167,9 +158,174 @@ static frame_handler_t frame_handlers[] =
 	{ "TRK", "TRCK", id3v2_parse_TRCK }, /* track */
 	{ "TP1", "TPE1", id3v2_parse_TPE1 }, /* artist */
 	{ "TYE", "TYER", id3v2_parse_TYER }, /* year */
-	{ NULL, NULL, NULL }
+	{ "", "", NULL }
 };
 
+/*****************************************************************************/
+/* ID3v1 support */
+
+static struct id3v1 *id3v1_alloc (struct id3_data *id3)
+{
+	struct id3v1 *tag;
+
+	if (!(tag = MALLOC (sizeof (struct id3v1))))
+		return NULL;
+
+	tag->id3 = id3;
+	return tag;
+}
+
+static void id3v1_free (struct id3v1 *tag)
+{
+	if (!tag)
+		return;
+
+	if (ID3_DATA(tag) && ID3_DATA(tag)->id3v1 == tag)
+		ID3_DATA(tag)->id3v1 = NULL;
+
+	free (tag);
+}
+
+static char *v1_str (unsigned char *p, size_t len)
+{
+	static char   v1_buf[64];
+	size_t        i;
+
+	for (i = 0; i < len; i++)
+	{
+		unsigned char c = p[i];
+
+		/* check if it was terminated, and i think 0xff is not valid */
+		if (c == 0 || c == 0xff)
+			break;
+
+		v1_buf[i] = p[i];
+	}
+
+	v1_buf[i] = 0;
+	string_trim (v1_buf);
+
+	return v1_buf;
+}
+
+/* The ID3v1 tag is always 128 bytes long and somewhere before
+ * the end of the file */
+static void read_v1 (struct id3v1 *tag)
+{
+	char            buf[512];
+	unsigned char  *p;
+	size_t          comment_len;
+	char           *str;
+
+	assert (ID3_DATA(tag)->f != NULL);
+
+	if (fseek (ID3_DATA(tag)->f, -128L, SEEK_END) != 0)
+		return;
+
+	if (fread (buf, 1, 128, ID3_DATA(tag)->f) != 128)
+		return;
+
+	if (strncmp (buf, "TAG", 3) != 0)
+		return;
+
+	p = buf + 3;
+
+	set_meta (ID3_DATA(tag), NULL, "title",   v1_str (p, 30)); p += 30;
+	set_meta (ID3_DATA(tag), NULL, "artist",  v1_str (p, 30)); p += 30;
+	set_meta (ID3_DATA(tag), NULL, "album",   v1_str (p, 30)); p += 30;
+	set_meta (ID3_DATA(tag), NULL, "year",    v1_str (p, 4));  p += 4;
+
+	/* check for track number, which is indicated by having byte
+	 * 29 of the comment field nulled (and a sane track number) */
+	if (p[28] == 0 && p[29] > 0 && p[29] != 0xff)
+		comment_len = 29;
+	else
+		comment_len = 30;
+
+	set_meta (ID3_DATA(tag), NULL, "comment", v1_str (p, comment_len));
+	p += comment_len;
+
+	if (comment_len == 29)
+	{
+		str = stringf_dup ("%d", *p);
+		set_meta (ID3_DATA(tag), NULL, "track", str);
+		free (str);
+		p++;
+	}
+
+	/* TODO: genre table? */
+}
+
+/*****************************************************************************/
+/* ID3v2 support */
+
+struct id3v2 *id3v2_alloc (struct id3_data *id3, size_t max_len)
+{
+	struct id3v2 *tag;
+
+	if (!(tag = MALLOC (sizeof (struct id3v2))))
+		return NULL;
+
+	tag->id3 = id3;
+
+	/* set the maximum frame size to read at any one time */
+	if (max_len == 0)
+		max_len = DEFAULT_MAX_FRAME_LEN;
+
+	tag->max_len = max_len;
+	return tag;
+}
+
+void id3v2_free (struct id3v2 *tag)
+{
+	if (!tag)
+		return;
+
+	if (ID3_DATA(tag) && ID3_DATA(tag)->id3v2 == tag)
+		ID3_DATA(tag)->id3v2 = NULL;
+
+	/* frames aren't stored, so there's nothing else to free */
+	free (tag);
+}
+
+static struct id3v2_frame *id3v2_frame_alloc (struct id3v2 *tag, const char *id)
+{
+	struct id3v2_frame *frame;
+
+	if (!(frame = MALLOC (sizeof (struct id3v2_frame))))
+		return NULL;
+
+	frame->tag      = tag;
+	frame->contents = NULL;
+	frame->size     = 0;
+	frame->offset   = 0;
+
+	strcpy (frame->id, id);
+
+	return frame;
+}
+
+static void id3v2_frame_free (struct id3v2_frame *frame)
+{
+	if (!frame)
+		return;
+
+	free (frame->contents);
+	free (frame);
+}
+
+static int id3v2_major_version (struct id3v2 *tag)
+{
+	return (tag->version & 0xff00) >> 8;
+}
+
+#ifdef DEBUG_META_TAG /* XXX */
+static int id3v2_minor_version (struct id3v2 *tag)
+{
+	return (tag->version & 0x00ff);
+}
+#endif
+
 /*****************************************************************************/
 
 /* get bytes in big-endian ordering */
@@ -209,209 +365,16 @@ static size_t get_sync_safe (unsigned char **ptr, size_t len)
 	return decoded;
 }
 
-/* this will be used in scanning for lyrics3 and musicmatch tags */
-#if 0
-/* look for a particular byte pattern in the file */
-static off_t scan_signature (struct tag_file *tag, char *signature, off_t start)
-{
-	int    c;
-	char  *p;
-	off_t  pos;
-
-	if (fseek (tag->f, start, SEEK_SET) != 0)
-		return -1;
-
-	pos = start;
-	p   = signature;
-
-	while (*p && (c = fgetc (tag->f)) != EOF)
-	{
-		pos++;
-
-		if (c == *p)
-			p++;
-		else
-			p = signature;
-	}
-
-	/* if we reached the end of the signature we found it */
-	return (*p == 0 ? pos - 3 : -1);
-}
-#endif
-
-/*****************************************************************************/
-/* ID3v1 support */
-
-static struct id3v1 *id3v1_alloc (struct tag_file *file)
-{
-	struct id3v1 *tag;
-
-	if (!(tag = MALLOC (sizeof (struct id3v1))))
-		return NULL;
-
-	tag->file = file;
-	return tag;
-}
-
-static void id3v1_free (struct id3v1 *tag)
-{
-	if (!tag)
-		return;
-
-	if (TAG_FILE(tag) && TAG_FILE(tag)->id3v1 == tag)
-		TAG_FILE(tag)->id3v1 = NULL;
-
-	free (tag);
-}
-
-static char *v1_str (unsigned char *p, size_t len)
-{
-	static char   v1_buf[64];
-	size_t        i;
-
-	for (i = 0; i < len; i++)
-	{
-		unsigned char c = p[i];
-
-		/* check if it was terminated, and i think 0xff is not valid */
-		if (c == 0 || c == 0xff)
-			break;
-
-		v1_buf[i] = p[i];
-	}
-
-	v1_buf[i] = 0;
-	string_trim (v1_buf);
-
-	return v1_buf;
-}
-
-/* The ID3v1 tag is always 128 bytes long and somewhere before
- * the end of the file */
-static void read_v1 (struct id3v1 *tag)
-{
-	char            buf[512];
-	unsigned char  *p;
-	size_t          comment_len;
-	char           *str;
-
-	if (fread (buf, 1, 128, TAG_FILE(tag)->f) != 128)
-		return;
-
-	if (strncmp (buf, "TAG", 3) != 0)
-		return;
-
-	/* if the first char is not printable or empty, assume a bogus TAG */
-	p = buf + 3;
-	if (!isprint (*p) && *p != 0)
-		return;
-
-	tag->start = ftell (TAG_FILE(tag)->f) - 128;
-
-	set_meta (TAG_FILE(tag), NULL, "title",   v1_str (p, 30)); p += 30;
-	set_meta (TAG_FILE(tag), NULL, "artist",  v1_str (p, 30)); p += 30;
-	set_meta (TAG_FILE(tag), NULL, "album",   v1_str (p, 30)); p += 30;
-	set_meta (TAG_FILE(tag), NULL, "year",    v1_str (p, 4));  p += 4;
-
-	/* check for track number, which is indicated by having byte
-	 * 29 of the comment field nulled (and a sane track number) */
-	if (p[28] == 0 && p[29] > 0 && p[29] != 0xff)
-		comment_len = 29;
-	else
-		comment_len = 30;
-
-	set_meta (TAG_FILE(tag), NULL, "comment", v1_str (p, comment_len));
-	p += comment_len;
-
-	if (comment_len == 29)
-	{
-		str = stringf_dup ("%d", *p);
-		set_meta (TAG_FILE(tag), NULL, "track", str);
-		free (str);
-		p++;
-	}
-
-	/* TODO: genre table? */
-}
-
-/*****************************************************************************/
-/* ID3v2 support */
-
-struct id3v2 *id3v2_alloc (struct tag_file *file, size_t max_len)
-{
-	struct id3v2 *tag;
-
-	if (!(tag = MALLOC (sizeof (struct id3v2))))
-		return NULL;
-
-	tag->file = file;
-
-	/* set the maximum frame size to read at any one time */
-	if (max_len == 0)
-		max_len = DEFAULT_MAX_FRAME_LEN;
-
-	tag->max_len = max_len;
-	return tag;
-}
-
-void id3v2_free (struct id3v2 *tag)
-{
-	if (!tag)
-		return;
-
-	if (TAG_FILE(tag) && TAG_FILE(tag)->id3v2 == tag)
-		TAG_FILE(tag)->id3v2 = NULL;
-
-	/* frames aren't stored, so there's nothing else to free */
-	free (tag);
-}
-
-static struct id3v2_frame *id3v2_frame_alloc (struct id3v2 *tag, char *id)
-{
-	struct id3v2_frame *frame;
-
-	if (!(frame = MALLOC (sizeof (struct id3v2_frame))))
-		return NULL;
-
-	frame->tag = tag;
-	strcpy (frame->id, id);
-
-	return frame;
-}
-
-static void id3v2_frame_free (struct id3v2_frame *frame)
-{
-	if (!frame)
-		return;
-
-	free (frame->contents);
-	free (frame);
-}
-
-static int id3v2_major_version (struct id3v2 *tag)
-{
-	return (tag->version & 0xff00) >> 8;
-}
-
-#ifdef DEBUG_META_TAG /* XXX */
-static int id3v2_minor_version (struct id3v2 *tag)
-{
-	return (tag->version & 0x00ff);
-}
-#endif
-
-/*****************************************************************************/
-
 /*
  * Read some data from the tag, reverse any unsynchronization replacing
  * sequences of [ff 00] with [ff], and copy the data to a buffer.
  *
  * NOTE: This method doesn't unsync sequences straddling the boundaries of
- * reads, so each read_resync() _must_ completely encompass all possible
+ * reads, so each id3v2_resync() _must_ completely encompass all possible
  * unsynchronized sequences within a given subrange of the tag. Otherwise
  * the tag will not be read correctly.
  */
-static size_t read_resync (struct id3v2 *tag, char *buf, size_t len)
+static size_t id3v2_resync (struct id3v2 *tag, char *buf, size_t len)
 {
 	int     c;
 	int     cprev;
@@ -425,7 +388,7 @@ static size_t read_resync (struct id3v2 *tag, char *buf, size_t len)
 
 	while (tag->data_offset < tag->size && n < len)
 	{
-		if ((c = fgetc (TAG_FILE(tag)->f)) == EOF)
+		if ((c = getc (ID3_DATA(tag)->f)) == EOF)
 			break;
 
 		tag->data_offset++;
@@ -457,7 +420,7 @@ static size_t read_resync (struct id3v2 *tag, char *buf, size_t len)
 
 /* read some data from the tag, possibly unsyncing it, into the supplied
  * buffer and update tag->data_offset */
-static size_t id3_read (struct id3v2 *tag, void *buf, size_t need_len)
+static size_t id3v2_read (struct id3v2 *tag, void *buf, size_t need_len)
 {
 	size_t unsynced_len;
 
@@ -467,7 +430,7 @@ static size_t id3_read (struct id3v2 *tag, void *buf, size_t need_len)
 		 * If the tag does not need unsynchronization, or if it
 		 * needs unsync but is version 4, we can just do fread().
 		 */
-		unsynced_len = fread (buf, 1, need_len, TAG_FILE(tag)->f);
+		unsynced_len = fread (buf, 1, need_len, ID3_DATA(tag)->f);
 
 		tag->data_offset += unsynced_len;
 		tag->fake_offset += unsynced_len;
@@ -479,9 +442,9 @@ static size_t id3_read (struct id3v2 *tag, void *buf, size_t need_len)
 	 * The tag needs unsynchronization, and we've got to do it the
 	 * hard way: one character at a time.
 	 *
-	 * tag->data_offset is handled by read_resync().
+	 * tag->data_offset is handled by id3v2_resync().
 	 */
-	unsynced_len = read_resync (tag, (char *)buf, need_len);
+	unsynced_len = id3v2_resync (tag, (char *)buf, need_len);
 	return unsynced_len;
 }
 
@@ -489,12 +452,12 @@ static size_t id3_read (struct id3v2 *tag, void *buf, size_t need_len)
  * Seek to a particular spot in the tag.
  *
  * If the tag is unsynchronized and the id3 major version is 2 or 3, we
- * have to fake doing a seek by doing id3_read(). We do this because in
+ * have to fake doing a seek by doing id3v2_read(). We do this because in
  * those versions of the standard, all sizes in the tag except the tag size
  * are _logical_ sizes, and the real sizes can only be obtained after
  * unsyncing the whole tag.
  */
-static int id3_seek (struct id3v2 *tag, size_t offset, int whence)
+static int id3v2_seek (struct id3v2 *tag, size_t offset, int whence)
 {
 	size_t ret;
 	size_t len;
@@ -503,26 +466,26 @@ static int id3_seek (struct id3v2 *tag, size_t offset, int whence)
 
 	if (!tag->unsynched || id3v2_major_version (tag) >= 4)
 	{
-		if ((ret = fseek (TAG_FILE(tag)->f, (long)tag->start + offset,
-		                  whence)) == 0)
-		{
-			tag->data_offset = offset;
-			tag->fake_offset = offset;
-		}
+		if (offset > LONG_MAX)
+			return -1;
 
-		return ret;
+		if (fseek (ID3_DATA(tag)->f, (long)offset, SEEK_SET) != 0)
+			return -1;
+
+		/* update real and virtual offsets, which are the same when the tag is
+		 * not unsynchronized */
+		tag->data_offset = offset;
+		tag->fake_offset = offset;
+
+		return 0;
 	}
 
-	/* only do this for tags at the start of the file */
-	if (tag->start != 0)
-		return -1;
-
 	/* can't seek backwards */
 	if (tag->fake_offset > offset)
 		return -1;
 
 	len = offset - tag->fake_offset;
-	ret = id3_read (tag, NULL, len);
+	ret = id3v2_read (tag, NULL, len);
 
 	if (ret != len)
 		return -1;
@@ -531,20 +494,19 @@ static int id3_seek (struct id3v2 *tag, size_t offset, int whence)
 }
 
 /* UNTESTED */
-/* TODO: limit extended header size to a sane limit */
 static BOOL skip_ext_header (struct id3v2 *tag, size_t start)
 {
 	unsigned char    buf[ID3V2_EXT_HDR_LEN];
 	unsigned char   *p;
 	size_t           newpos;
 
-	if (!TAG_FILE(tag)->f)
+	if (!ID3_DATA(tag)->f)
 		return FALSE;
 
-	if (id3_seek (tag, start, SEEK_SET) != 0)
+	if (id3v2_seek (tag, start, SEEK_SET) != 0)
 		return FALSE;
 
-	if (id3_read (tag, buf, ID3V2_EXT_HDR_LEN) != ID3V2_EXT_HDR_LEN)
+	if (id3v2_read (tag, buf, ID3V2_EXT_HDR_LEN) != ID3V2_EXT_HDR_LEN)
 		return FALSE;
 
 	p = buf;
@@ -558,6 +520,10 @@ static BOOL skip_ext_header (struct id3v2 *tag, size_t start)
 	else
 		tag->ext_hdr_len = get_bytes (&p, 4);
 
+	/* the extended header can't be too big */
+	if (tag->ext_hdr_len > MAX_FRAME_SIZE)
+		return FALSE;
+
 	newpos = start + tag->ext_hdr_len;
 
 	/*
@@ -575,7 +541,7 @@ static BOOL skip_ext_header (struct id3v2 *tag, size_t start)
 	 * Skip the header, as it doesn't have any frames we care
 	 * about at the moment (crc32, update flag, ...)
 	 */
-	if (id3_seek (tag, newpos, SEEK_SET) != 0)
+	if (id3v2_seek (tag, newpos, SEEK_SET) != 0)
 		return FALSE;
 
 	return TRUE;
@@ -624,7 +590,7 @@ static BOOL parse_v2_header (struct id3v2 *tag, unsigned char *hdr,
 	int              not_found;
 	unsigned char   *h;
 	size_t           sig_len;
-	unsigned int     flags;
+	unsigned char    flags;
 
 	h = hdr;
 	not_found = 0;
@@ -650,7 +616,7 @@ static BOOL parse_v2_header (struct id3v2 *tag, unsigned char *hdr,
 	if (not_found)
 		return FALSE;
 
-	META_TAG_TRACE (("---: [%s] found \"%s\"\n", TAG_FILE(tag)->name,
+	META_TAG_TRACE (("---: [%s] found \"%s\"\n", ID3_DATA(tag)->name,
 	                 signature));
 
 	h = hdr + sig_len;
@@ -677,7 +643,7 @@ static BOOL parse_v2_header (struct id3v2 *tag, unsigned char *hdr,
 	set_tag_flags (tag, flags);
 
 	META_TAG_TRACE (("[%s]: unsync_tag: %i extended: %i experimental: %i "
-	                 "has_footer: %i\n", TAG_FILE(tag)->name, tag->unsynched,
+	                 "has_footer: %i\n", ID3_DATA(tag)->name, tag->unsynched,
 	                 tag->has_ext_hdr, tag->experimental, tag->has_footer));
 
 	/*
@@ -697,48 +663,22 @@ static BOOL parse_v2_header (struct id3v2 *tag, unsigned char *hdr,
 static void read_v2 (struct id3v2 *tag, char *signature)
 {
 	unsigned char    header[ID3V2_HDR_LEN];
-	off_t            pos;
 
-	if (!TAG_FILE(tag)->f)
+	assert (ID3_DATA(tag)->f != NULL);
+
+	if (fseek (ID3_DATA(tag)->f, 0, SEEK_SET) != 0)
 		return;
 
-	if (fread (header, 1, ID3V2_HDR_LEN, TAG_FILE(tag)->f) != ID3V2_HDR_LEN)
+	if (fread (header, 1, ID3V2_HDR_LEN, ID3_DATA(tag)->f) != ID3V2_HDR_LEN)
 		return;
 
 	if (!parse_v2_header (tag, header, signature))
 		return;
 
-	if ((pos = ftell (TAG_FILE(tag)->f)) == -1)
-		return;
-
-	/* store the tag start position */
-	tag->start = pos - ID3V2_HDR_LEN;
-
 	/* put the real and virtual offsets after the header */
 	tag->data_offset = ID3V2_HDR_LEN;
 	tag->fake_offset = ID3V2_HDR_LEN;
 
-	/* only accept version 4 tags anywhere but the start */
-	if (tag->start != 0 && id3v2_major_version (tag) != 4)
-		return;
-
-	/*
-	 * If this was a footer we read (indicated by looking for
-	 * backwards "ID3"), then restart the tag search at the
-	 * beginning of the tag.
-	 */
-	if (!strncmp (signature, "3DI", 3))
-	{
-		off_t newpos = pos - tag->size;
-
-		if (fseek (TAG_FILE(tag)->f, newpos, SEEK_SET) != 0)
-			return;
-
-		/* restart the read at the new position */
-		read_v2 (tag, "ID3");
-		return;
-	}
-
 	/* can't handle compressed tags yet */
 	if (tag->compressed)
 		return;
@@ -770,11 +710,11 @@ static size_t frame_id_len (struct id3v2 *tag)
 	return 4;
 }
 
-static BOOL is_frameid_char (char c)
+static BOOL is_frameid_char (int c)
 {
-	if (c >= 'A' && c <= 'Z')
+	if (c >= 65 /*'A'*/ && c <= 90 /*'Z'*/)
 		return TRUE;
-	else if (c >= '0' && c <= '9')
+	else if (c >= 48 /*'0'*/ && c <= 57 /*'9'*/)
 		return TRUE;
 
 	return FALSE;
@@ -822,18 +762,19 @@ static void set_frame_flags (struct id3v2_frame *frame, int flags1, int flags2)
 	frame->has_data_len = (flags2 & 0x01) != 0;
 }
 
-static BOOL frame_resize_buffer (struct id3v2_frame *frame, size_t len)
+static BOOL frame_alloc_buffer (struct id3v2_frame *frame, size_t len)
 {
-	unsigned char *newbuf;
+	uint8_t *buf;
 
-	if (len <= frame->size)
-		return TRUE;
+	assert (len > 0);
+	assert (len <= MAX_FRAME_SIZE + 1);
+	assert (frame->contents == NULL);
 
-	if (!(newbuf = realloc (frame->contents, len)))
+	if (!(buf = malloc (len)))
 		return FALSE;
 
 	frame->size     = len;
-	frame->contents = newbuf;
+	frame->contents = buf;
 
 	return TRUE;
 }
@@ -858,33 +799,32 @@ static frame_handler_t *find_handler (frame_handler_t *p, char *id)
 static BOOL read_frame (struct id3v2 *tag, struct id3v2_frame *frame,
                         frame_handler_t *handler, size_t len)
 {
-	ssize_t  n;
+	size_t n;
 
 	/* can't handle compressed or encrypted frames */
 	if (!handler || frame->encrypted || frame->compressed)
 		return FALSE;
 
-	/* check if this frame would exceed memory allotment
-	 * (+1 to allow for null terminator) */
-	if (len + 1 > tag->max_len)
+	/* check if this frame would exceed memory allotment */
+	if (len > tag->max_len)
 	{
 		META_TAG_TRACE (("%s EXCEEDS max len (%u > %u)\n",
 		                 frame->id, len, tag->max_len));
 		return FALSE;
 	}
 
-	if (!frame_resize_buffer (frame, len + 1))
+	if (!frame_alloc_buffer (frame, len + 1))
 		return FALSE;
 
 	/*
 	 * Version 4 of the standard allows for avoiding all the unsync
 	 * problems, but we have to unsync the frame contents here
-	 * instead of dynamically in id3_read().
+	 * instead of dynamically in id3v2_read().
 	 */
 	if ((tag->unsynched || frame->unsynched) && id3v2_major_version (tag) >= 4)
-		n = read_resync (tag, frame->contents, len);
+		n = id3v2_resync (tag, frame->contents, len);
 	else
-		n = id3_read (tag, frame->contents, len);
+		n = id3v2_read (tag, frame->contents, len);
 
 	if (n != len)
 	{
@@ -901,6 +841,26 @@ static BOOL read_frame (struct id3v2 *tag, struct id3v2_frame *frame,
 	return TRUE;
 }
 
+static size_t get_frame_len (struct id3v2 *tag, unsigned char *p)
+{
+	size_t data_len;
+
+	/*
+	 * v2.2.0 - 3 bytes length field
+	 * v2.3.0 - 4 bytes ""
+	 * v2.4.0 - sync-safe
+	 */
+	switch (id3v2_major_version (tag))
+	{
+	 case 2: data_len = get_bytes     (&p, 3); break;
+	 case 3: data_len = get_bytes     (&p, 4); break;
+	 case 4: data_len = get_sync_safe (&p, 4); break;
+	 default: abort ();
+	}
+
+	return data_len;
+}
+
 static void read_v2_frames (struct id3v2 *tag)
 {
 	unsigned char        buf[ID3V2_FRAME_HDR_MAX];
@@ -910,7 +870,7 @@ static void read_v2_frames (struct id3v2 *tag)
 	size_t               pos;
 	struct id3v2_frame  *frame;
 
-	if (!TAG_FILE(tag)->f)
+	if (!ID3_DATA(tag)->f)
 		return;
 
 	id_len  = frame_id_len  (tag);
@@ -929,14 +889,14 @@ static void read_v2_frames (struct id3v2 *tag)
 		int               i;
 		frame_handler_t  *handler;
 
-		if (id3_seek (tag, pos, SEEK_SET) != 0)
+		if (id3v2_seek (tag, pos, SEEK_SET) != 0)
 			return;
 
 		if (tag->data_offset >= tag->size)
 			break;
 
 		/* read the frame header */
-		if (id3_read (tag, buf, hdr_len) != hdr_len)
+		if (id3v2_read (tag, buf, hdr_len) != hdr_len)
 			return;
 
 		memcpy (frame_id, buf, id_len);
@@ -944,23 +904,17 @@ static void read_v2_frames (struct id3v2 *tag)
 
 		/* check for a frame id (possibly padding) without ascii characters */
 		for (i = 0; i < id_len; i++)
+		{
 			if (!is_frameid_char (frame_id[i]))
 			    return;
+		}
 
 		p = buf + id_len;
+		data_len = get_frame_len (tag, p);
 
-		/*
-		 * v2.2.0 - 3 bytes length field
-		 * v2.3.0 - 4 bytes ""
-		 * v2.4.0 - sync-safe
-		 */
-		switch (id3v2_major_version (tag))
-		{
-			case 2: data_len = get_bytes     (&p, 3); break;
-			case 3: data_len = get_bytes     (&p, 4); break;
-			case 4: data_len = get_sync_safe (&p, 4); break;
-			default: assert (0);
-		}
+		/* sanity check the length */
+		if (data_len > MAX_FRAME_SIZE)
+			break;
 
 		/*
 		 * Could try to continue parsing here. The most likely occurrence
@@ -982,7 +936,7 @@ static void read_v2_frames (struct id3v2 *tag)
 		set_frame_flags (frame, buf[8], buf[9]);
 
 		META_TAG_TRACE (("###: %s LEN: %u [%lu]\n", frame_id, data_len,
-		                 ftell (TAG_FILE(tag)->f)));
+		                 ftell (ID3_DATA(tag)->f)));
 
 		META_TAG_TRACE (("%s: group: %i compress: %i encrypt: %i "
 		                 "unsync: %i data_len: %i\n", frame->id,
@@ -1022,18 +976,31 @@ static char *id3v2_frame_get_str (struct id3v2_frame *frame)
 	return p;
 }
 
-static char *id3v2_frame_get_str_with_encoding (struct id3v2_frame *frame)
+static char *id3v2_frame_get_encoded_str (struct id3v2_frame *frame)
 {
 	uint8_t encoding = id3v2_frame_get_uint8 (frame);
 
-	/* only support ISO-8859-1 encoding (marked with zero byte)
-	 * NOTE: this assumes that is the default encoding being used */
+	/*
+	 * Only support ISO-8859-1 encoding, which is marked with zero byte in the
+	 * "encoding" byte.
+	 *
+	 * TODO: This is wrong, as only UTF-8 metadata should exist from giftd's
+	 * and each plugin's point of view.
+	 */
 	if (encoding != 0)
 		return NULL;
 
 	return id3v2_frame_get_str (frame);
 }
 
+static BOOL is_lang_char (int c)
+{
+	if (c >= 97 /*'a'*/ && c <= 122 /*'z'*/)
+		return TRUE;
+
+	return FALSE;
+}
+
 static char *id3v2_frame_get_lang (struct id3v2_frame *frame)
 {
 	static char lang[4];
@@ -1054,7 +1021,7 @@ static char *id3v2_frame_get_lang (struct id3v2_frame *frame)
 		if (!(lang[i] = id3v2_frame_get_uint8 (frame)))
 		    break;
 
-		if (!isalnum (lang[i]))
+		if (!is_lang_char (lang[i]))
 			break;
 	}
 
@@ -1071,21 +1038,21 @@ static char *id3v2_frame_get_lang (struct id3v2_frame *frame)
 
 static void id3v2_parse_TLEN (struct id3v2 *tag, struct id3v2_frame *frame)
 {
-	int    millis;
-	int    n;
-	char  *p;
-	char   buf[128];
+	unsigned int millis;
+	int          n;
+	char        *str;
+	char         buf[128];
 
-	if (!(p = id3v2_frame_get_str_with_encoding (frame)))
+	if (!(str = id3v2_frame_get_encoded_str (frame)))
 		return;
 
-	META_TAG_TRACE (("TLEN: contents=%s [len=%d]\n", p, frame->size));
+	META_TAG_TRACE (("TLEN: contents=%s [len=%d]\n", str, frame->size));
 
-	if ((n = sscanf (p, "%d", &millis)) != 1)
+	if ((n = sscanf (str, "%u", &millis)) != 1)
 		return;
 
 	snprintf (buf, sizeof (buf) - 1, "%u", millis / 1000);
-	set_meta (TAG_FILE(tag), frame->id, "duration", buf);
+	set_meta (ID3_DATA(tag), frame->id, "duration", buf);
 }
 
 static void id3v2_parse_COMM (struct id3v2 *tag, struct id3v2_frame *frame)
@@ -1101,165 +1068,105 @@ static void id3v2_parse_COMM (struct id3v2 *tag, struct id3v2_frame *frame)
 	META_TAG_TRACE (("COMM: lang=%s description=%s\n", lang, descr));
 	lang = descr; /* prevent 'unused variable' warnings */
 
-	set_meta (TAG_FILE(tag), frame->id, "comment",
+	set_meta (ID3_DATA(tag), frame->id, "comment",
 	          id3v2_frame_get_str (frame));
 }
 
 static void id3v2_parse_TIT2 (struct id3v2 *tag, struct id3v2_frame *frame)
 {
-	set_meta (TAG_FILE(tag), frame->id, "title",
-	          id3v2_frame_get_str_with_encoding (frame));
+	set_meta (ID3_DATA(tag), frame->id, "title",
+	          id3v2_frame_get_encoded_str (frame));
 }
 
 static void id3v2_parse_TALB (struct id3v2 *tag, struct id3v2_frame *frame)
 {
-	set_meta (TAG_FILE(tag), frame->id, "album",
-	          id3v2_frame_get_str_with_encoding (frame));
+	set_meta (ID3_DATA(tag), frame->id, "album",
+	          id3v2_frame_get_encoded_str (frame));
 }
 
 static void id3v2_parse_TRCK (struct id3v2 *tag, struct id3v2_frame *frame)
 {
-	set_meta (TAG_FILE(tag), frame->id, "track",
-	          id3v2_frame_get_str_with_encoding (frame));
+	set_meta (ID3_DATA(tag), frame->id, "track",
+	          id3v2_frame_get_encoded_str (frame));
 }
 
 static void id3v2_parse_TPE1 (struct id3v2 *tag, struct id3v2_frame *frame)
 {
-	set_meta (TAG_FILE(tag), frame->id, "artist",
-	          id3v2_frame_get_str_with_encoding (frame));
+	set_meta (ID3_DATA(tag), frame->id, "artist",
+	          id3v2_frame_get_encoded_str (frame));
 }
 
 static void id3v2_parse_TYER (struct id3v2 *tag, struct id3v2_frame *frame)
 {
-	set_meta (TAG_FILE(tag), frame->id, "year",
-	          id3v2_frame_get_str_with_encoding (frame));
+	set_meta (ID3_DATA(tag), frame->id, "year",
+	          id3v2_frame_get_encoded_str (frame));
 }
 
 /*****************************************************************************/
 
-static void tag_file_close (struct tag_file *tagfile)
+static void id3_data_close (struct id3_data *id3)
 {
-	if (!tagfile)
+	if (!id3)
 		return;
 
-	free (tagfile->name);
+	free (id3->name);
 
-	if (tagfile->f)
+	if (id3->f)
 	{
-		fclose (tagfile->f);
-		tagfile->f = NULL;
+		fclose (id3->f);
+		id3->f = NULL;
 	}
 
-	id3v2_free (tagfile->id3v2);
-	id3v1_free (tagfile->id3v1);
+	id3v2_free (id3->id3v2);
+	id3v1_free (id3->id3v1);
 
-	dataset_clear (tagfile->keys);
-	tagfile->keys = NULL;
+	dataset_clear (id3->keys);
+	id3->keys = NULL;
 
-	free (tagfile);
+	free (id3);
 }
 
-static struct tag_file *tag_file_open (Share *share, const char *path,
+static struct id3_data *id3_data_open (Share *share, const char *path,
                                        size_t max_len)
 {
-	struct tag_file *file;
-	struct stat      st;
+	struct id3_data *id3;
 
-	if (stat (path, &st) != 0)
+	if (!(id3 = malloc (sizeof (struct id3_data))))
 		return NULL;
 
-	if (!(file = MALLOC (sizeof (struct tag_file))))
-		return NULL;
+	id3->f     = NULL;
+	id3->name  = STRDUP (path);
+	id3->keys  = NULL;
+	id3->share = share;
 
-	file->id3v2 = id3v2_alloc (file, max_len);
-	file->id3v1 = id3v1_alloc (file);
+	id3->id3v1 = id3v1_alloc (id3);
+	id3->id3v2 = id3v2_alloc (id3, max_len);
 
-	if (!file->id3v1 || !file->id3v2)
+	if (!id3->id3v1 || !id3->id3v2)
 	{
-		tag_file_close (file);
+		id3_data_close (id3);
 		return NULL;
 	}
 
-	if (!(file->f = fopen (path, "rb")))
+	if (!(id3->f = fopen (path, "rb")))
 	{
-		tag_file_close (file);
+		id3_data_close (id3);
 		return NULL;
 	}
 
-	file->name  = STRDUP (file_basename ((char *)path));
-	file->size  = st.st_size;
-	file->share = share;
-
-	return file;
+	return id3;
 }
 
-static void check_id3v1 (struct id3v1 *tag, off_t pos)
+static void id3_data_read (struct id3_data *id3)
 {
-	if (fseek (TAG_FILE(tag)->f, pos, SEEK_SET) != 0)
-		return;
-
-	read_v1 (tag);
-}
-
-static void check_id3v2 (struct id3v2 *tag, char *signature, off_t pos)
-{
-	if (fseek (TAG_FILE(tag)->f, pos, SEEK_SET) != 0)
-		return;
-
-	read_v2 (tag, signature);
-}
-
-static void tag_file_read (struct tag_file *file)
-{
-	off_t start;
-	off_t pos;
-
-	start = file->size - 4096;
-
-	if (file->size < 4096)
-		start = 0;
-
-	/* scan id3v1{,.1} tag first */
-	pos = file->size - 128;
-	check_id3v1 (file->id3v1, pos);
-
-	/* TODO */
-#if 0
-	/* check musicmatch tags */
-	pos = start;
-	while ((pos = scan_signature (file, MUSIC_MATCH_MAGIC)) >= 0)
-	{
-		check_musicmatch (file->musicmatch, pos);
-		pos += strlen (MUSIC_MATCH_MAGIC);
-	}
-
-	/* check lyrics3 tags */
-	pos = start;
-	if (!dataset_lookup (
-	while ((pos = scan_signature (file, "LYRICSBEGIN")) >= 0)
-	{
-		check_lyrics3 (file->lyrics3, pos);
-		pos += strlen ("LYRICSBEGIN");
-	}
-#endif
-
-	/* read id3v2 tag at the start of the file */
-	check_id3v2 (file->id3v2, "ID3", 0);
-
-	/* read id3v2 tags at the end, looking for footer signature "3DI" */
-	pos = file->size - ID3V2_HDR_LEN;
-
-	/* footer of id3v2 tag should be just before id3v1 tag */
-	if (file->id3v1->start > 0)
-		pos -= file->id3v1->start;
-
-	check_id3v2 (file->id3v2, "3DI", pos);
+	read_v1 (id3->id3v1);
+	read_v2 (id3->id3v2, "ID3");
 }
 
 /*****************************************************************************/
 
 /* set a meta key read from the file */
-static void set_meta (struct tag_file *file, char *id, char *key, char *val)
+static void set_meta (struct id3_data *id3, char *id, char *key, char *val)
 {
 	char   *old;
 	char   *max_str;
@@ -1270,7 +1177,7 @@ static void set_meta (struct tag_file *file, char *id, char *key, char *val)
 		return;
 
 	/* only handle certain keys */
-	if (!(max_str = dataset_lookupstr (file->keys, key)))
+	if (!(max_str = dataset_lookupstr (id3->keys, key)))
 		return;
 
 	max_len = ATOUL (max_str);
@@ -1278,7 +1185,7 @@ static void set_meta (struct tag_file *file, char *id, char *key, char *val)
 	/* trim the data */
 	string_trim (val);
 
-	old = share_get_meta (file->share, key);
+	old = share_get_meta (id3->share, key);
 	len = strlen (val);
 
 	/* if the old id was larger, take that one instead */
@@ -1299,39 +1206,39 @@ static void set_meta (struct tag_file *file, char *id, char *key, char *val)
 	else
 		META_TAG_TRACE (("=== [v1] (%s): %s\n", key, val));
 
-	share_set_meta (file->share, key, val);
+	share_set_meta (id3->share, key, val);
 }
 
-static void add_key (struct tag_file *file, char *key, size_t max_len)
+static void add_key (struct id3_data *id3, char *key, size_t max_len)
 {
 	char *str;
 
 	if (!(str = stringf_dup ("%lu", (unsigned long)max_len)))
 		return;
 
-	dataset_insertstr (&file->keys, key, str);
+	dataset_insertstr (&id3->keys, key, str);
 	free (str);
 }
 
 void meta_tag_run (Share *share, const char *path)
 {
-	struct tag_file *file;
+	struct id3_data *id3;
 
-	if (!(file = tag_file_open (share, path, DEFAULT_MAX_FRAME_LEN)))
+	if (!(id3 = id3_data_open (share, path, DEFAULT_MAX_FRAME_LEN)))
 		return;
 
 	/* insert the keys we are interested in, as well as the max length
 	 * for each key */
-	add_key (file, "duration", 128);
-	add_key (file, "comment",  128);
-	add_key (file, "title",    128);
-	add_key (file, "album",    128);
-	add_key (file, "track",    10);
-	add_key (file, "artist",   128);
-	add_key (file, "year",     6);
+	add_key (id3, "duration", 128);
+	add_key (id3, "comment",  128);
+	add_key (id3, "title",    128);
+	add_key (id3, "album",    128);
+	add_key (id3, "track",    10);
+	add_key (id3, "artist",   128);
+	add_key (id3, "year",     6);
 
-	tag_file_read (file);
-	tag_file_close (file);
+	id3_data_read (id3);
+	id3_data_close (id3);
 }
 
 /*****************************************************************************/
@@ -1339,7 +1246,7 @@ void meta_tag_run (Share *share, const char *path)
 #if 0
 int main (int argc, char **argv)
 {
-	struct tag_file *tags;
+	struct id3_data *tags;
 	Share            share;
 	int              i;
 
@@ -1348,11 +1255,11 @@ int main (int argc, char **argv)
 		printf ("checking %s\n", argv[i]);
 		share_init (&share);
 
-		if (!(tags = tag_file_open (&share, argv[i], 0)))
+		if (!(tags = id3_data_open (&share, argv[i], 0)))
 			continue;
 
-		tag_file_read (tags);
-		tag_file_close (tags);
+		id3_data_read (tags);
+		id3_data_close (tags);
 
 		share_finish (&share);
 	}
