@@ -1,5 +1,5 @@
 /*
- * $Id: share_cache.c,v 1.64 2003/10/31 13:14:48 jasta Exp $
+ * $Id: share_cache.c,v 1.71 2004/11/12 00:41:55 hexwab Exp $
  *
  * TODO: This file desperately needs to be rewritten.  Volunteers are
  * encouraged!
@@ -154,19 +154,16 @@ static List *build_share_root (void)
 	char *sharing_root, *sharing_root0;
 	char *root;
 
-	if (!(sharing_root = config_get_str (gift_conf, "sharing/root")))
+	if ((sharing_root = config_get_str (gift_conf, "sharing/root")))
 	{
-		GIFT_WARN (("no sharing root found!!  share you bastard!"));
-		return NULL;
+		sharing_root0 = sharing_root = STRDUP (sharing_root);
+
+		while ((root = string_sep (&sharing_root, ":")))
+			add_share_root (&sroot, file_expand_path (root), TRUE);
+
+		free (sharing_root0);
 	}
 
-	sharing_root0 = sharing_root = STRDUP (sharing_root);
-
-	while ((root = string_sep (&sharing_root, ":")))
-		add_share_root (&sroot, file_expand_path (root), TRUE);
-
-	free (sharing_root0);
-
 	if (SHARE_COMPLETED)
 	{
 		char *unix_path;
@@ -176,6 +173,12 @@ static List *build_share_root (void)
 		free (unix_path);
 	}
 
+	if (sroot == NULL)
+	{
+		GIFT_WARN (("no sharing root found!!  share you bastard!"));
+		return NULL;
+	}
+
 	return sroot;
 }
 
@@ -259,24 +262,6 @@ static FileShare *create_share_disk (char *host_root, char *host_path)
 
 /*****************************************************************************/
 
-static int cmp_list_entry (FileShare *file, char *path)
-{
-	if (!SHARE_DATA(file))
-		return -1;
-
-	if (!strcmp (SHARE_DATA(file)->path, path))
-		return 0;
-
-	return -1;
-}
-
-static int cmp_hash_entry (ds_data_t *key, ds_data_t *value, char *path)
-{
-	FileShare *file = value->data;
-
-	return (cmp_list_entry (file, path) == 0);
-}
-
 static FileShare *find_share_path (Dataset **shares, List **duplicates,
                                    char *path)
 {
@@ -286,35 +271,7 @@ static FileShare *find_share_path (Dataset **shares, List **duplicates,
 		return NULL;
 
 	if (shares)
-		file = dataset_find (*shares, DS_FIND(cmp_hash_entry), path);
-
-	if (!file && duplicates)
-	{
-		List *link;
-
-		link = list_find_custom (*duplicates, path,
-		                         (CompareFunc)cmp_list_entry);
-		file = list_nth_data (link, 0);
-	}
-
-	return file;
-}
-
-static FileShare *find_share_hash (Dataset **shares, List **duplicates,
-                                   Hash *hash)
-{
-	FileShare *file = NULL;
-
-	if (!hash)
-		return NULL;
-
-	if (shares)
-		file = dataset_lookup (*shares, hash->data, hash->len);
-
-#if 0
-	if (!file && duplicates)
-		/* TODO */;
-#endif
+		file = dataset_lookupstr (*shares, path);
 
 	return file;
 }
@@ -323,26 +280,13 @@ static FileShare *find_share_hash (Dataset **shares, List **duplicates,
 
 static void add_dataset_share (Dataset **dataset, FileShare *file)
 {
-	Hash *hash;
-
 	if (!dataset)
 		return;
 
-	if (!(hash = share_get_hash (file, NULL)))
-		return;
-
 	if (!(*dataset))
 		*dataset = dataset_new (DATASET_HASH);
 
-	dataset_insert (dataset, hash->data, hash->len, file, 0);
-}
-
-static void add_list_share (List **list, FileShare *file)
-{
-	if (!list)
-		return;
-
-	*list = list_prepend (*list, file);
+	dataset_insert (dataset, file->path, STRLEN_0(file->path), file, 0);
 }
 
 /* adds the share to the appropriate structure
@@ -350,54 +294,27 @@ static void add_list_share (List **list, FileShare *file)
  * shares */
 static int add_share (Dataset **shares, List **duplicates, FileShare *file)
 {
-	FileShare *cache = NULL;
-	Hash      *hash;
-
 	if (!share_complete (file))
 		return FALSE;
 
-	/*
-	 * This may return null, if the file only has currently unregistered
-	 * hashes. If so, add it to the list of duplicates temporarily, in
-	 * hopes of a registered algorithm coming along soon.
-	 */
-	hash = share_get_hash (file, NULL);
+	add_dataset_share (shares, file);
 
-	if (hash && !(cache = find_share_hash (shares, NULL, hash)))
-		add_dataset_share (shares, file);
-	else
-		add_list_share (duplicates, file);
-
-	return (cache == NULL);
+	return TRUE;
 }
 
 /*****************************************************************************/
 
 static void remove_dataset_share (Dataset **dataset, FileShare *file)
 {
-	Hash *hash;
-
 	if (!dataset)
 		return;
 
-	if (!(hash = share_get_hash (file, NULL)))
-		return;
-
-	dataset_remove (*dataset, hash->data, hash->len);
-}
-
-static void remove_list_share (List **duplicates, FileShare *file)
-{
-	if (!duplicates)
-		return;
-
-	*duplicates = list_remove (*duplicates, file);
+	dataset_removestr (*dataset, file->path);
 }
 
 static int remove_share (Dataset **shares, List **duplicates, FileShare *file)
 {
 	remove_dataset_share (shares, file);
-	remove_list_share (duplicates, file);
 	share_unref (file);
 	return TRUE;
 }
@@ -410,8 +327,6 @@ static BOOL handle_file (SDB *db, Dataset **shares, List **duplicates,
 {
 	FileShare *file  = NULL;
 	FileShare *cache = NULL;
-	FileShare *cache_s = NULL;
-	FileShare *cache_d = NULL;
 	char      *unix_path;
 	BOOL       usable;
 	BOOL       ret;
@@ -419,8 +334,7 @@ static BOOL handle_file (SDB *db, Dataset **shares, List **duplicates,
 	if (!(unix_path = file_unix_path (host_path)))
 		return FALSE;
 
-	if (!(cache_s = find_share_path (shares, NULL, unix_path)) &&
-	    !(cache_d = find_share_path (NULL, duplicates, unix_path)))
+	if (!(cache = find_share_path (shares, NULL, unix_path)))
 	{
 		if (!(file = create_share_disk (host_root, host_path)))
 		{
@@ -430,8 +344,6 @@ static BOOL handle_file (SDB *db, Dataset **shares, List **duplicates,
 	}
 	else
 	{
-		cache = (cache_s) ? cache_s : cache_d;
-
 		/* this path has been cached previously, if the mtimes differ remove
 		 * it from the list and recalculate it as if it's a new file */
 		if (SHARE_DATA(cache)->mtime != mtime)
@@ -449,7 +361,7 @@ static BOOL handle_file (SDB *db, Dataset **shares, List **duplicates,
 	if (!file)
 	{
 		file   = cache;
-		usable = BOOL_EXPR (cache_s != NULL);
+		usable = BOOL_EXPR (cache != NULL);
 	}
 	else
 	{
@@ -514,10 +426,42 @@ static int path_traverse (SDB *db, int total, SubprocessData *subproc,
 	char           newpath[PATH_MAX];
 	char          *path_fmt;
 	int            retval;
+	char          *incoming_dir;
 
 	if (!path)
 		return total;
 
+	/* don't share incoming dir if not explicitly allowed by config */
+	if (config_get_str (gift_conf, "sharing/ignore_incoming=1") &&
+	    (incoming_dir = STRDUP (INCOMING_PATH (("")))))
+	{
+		char *current_dir = STRDUP (path);
+		char *last_char = incoming_dir + strlen (incoming_dir) - 1;
+
+		/* remove trailing slash  */
+		if (*last_char == '/' || *last_char == '\\')
+			*last_char = 0;
+
+		/* current_dir must not be below incoming_dir, the opposite is ok */
+		if (strlen (current_dir) > strlen (incoming_dir))
+			current_dir[strlen (incoming_dir)] = 0;
+
+#ifndef WIN32
+		if (!strcmp (current_dir, incoming_dir))
+#else
+		if (!strcasecmp (current_dir, incoming_dir))
+#endif
+		{
+			GIFT_WARN (("Ignoring path below incoming dir: %s", path));
+			free (current_dir);
+			free (incoming_dir);
+			return total;
+		}
+
+		free (current_dir);
+		free (incoming_dir);
+	}
+
 	GIFT_TRACE (("descending %s...", path));
 
 	/* calculate the length here for optimization purposes */
@@ -540,18 +484,24 @@ static int path_traverse (SDB *db, int total, SubprocessData *subproc,
 		if (!d->d_name || !d->d_name[0])
 			continue;
 
-		if (HIDE_DOT_FILES)
-		{
-			if (d->d_name[0] == '.')
-				continue;
-		}
-
 		/* ignore '.' and '..' */
 		if (!strcmp (d->d_name, ".") || !strcmp (d->d_name, ".."))
 			continue;
 
 		snprintf (newpath, sizeof (newpath) - 1, path_fmt, path, d->d_name);
 
+		if (HIDE_DOT_FILES)
+		{
+			if (d->d_name[0] == '.')
+				continue;
+
+#ifdef WIN32
+			/* additionally check for hidden attribute on windows */
+			if (GetFileAttributes (newpath) & FILE_ATTRIBUTE_HIDDEN)
+				continue;
+#endif
+		}
+
 #ifndef WIN32
 		if (!FOLLOW_SYMLINKS)
 			retval = lstat (newpath, &st);
@@ -968,13 +918,7 @@ static void emit_share_new (ds_data_t *key, ds_data_t *value, FileShare *file)
 	Protocol *p = value->data;
 	void     *data;
 
-	/* make sure we can't have any data left over, this should probably
-	 * be an assert */
-	if ((data = share_get_udata (file, p->name)))
-	{
-		p->share_free (p, file, data);
-		share_set_udata (file, p->name, NULL);
-	}
+	assert (share_get_udata (file, p->name) == NULL);
 
 	/*
 	 * Ask the protocol to create new arbitrary data (if it needs to be
@@ -985,11 +929,13 @@ static void emit_share_new (ds_data_t *key, ds_data_t *value, FileShare *file)
 		share_set_udata (file, p->name, data);
 }
 
+#if 0
 static void emit_share_new_all (ds_data_t *key, ds_data_t *value, void *udata)
 {
 	/* we need to emit share_new for all protocols, as well */
 	plugin_foreach (DS_FOREACH(emit_share_new), value->data);
 }
+#endif
 
 static void emit_share_free (ds_data_t *key, ds_data_t *value, FileShare *file)
 {
@@ -1009,10 +955,12 @@ static void emit_share_free (ds_data_t *key, ds_data_t *value, FileShare *file)
 	}
 }
 
+#if 0
 static void emit_share_free_all (ds_data_t *key, ds_data_t *value, void *udata)
 {
 	plugin_foreach (DS_FOREACH(emit_share_free), value->data);
 }
+#endif
 
 static void emit_share_sync (ds_data_t *key, ds_data_t *value, int *begin)
 {
@@ -1030,33 +978,36 @@ static void emit_share_sync (ds_data_t *key, ds_data_t *value, int *begin)
 /*****************************************************************************/
 
 /*
- * Attempts to determine if data expressed by a_node exists in both dataset a
- * and b.  Returns TRUE if that is the case.
+ * Attempts to determine if data expressed by a_node exists in dataset b
+ * too.  Returns the node if that is the case.
  *
  * This is done so that we can determine the practical differences between
  * the old set of files shared and the new.
  */
-static int eq_filesets (ds_data_t *key, ds_data_t *value, Dataset *b)
+static DatasetNode *eq_filesets (ds_data_t *key, ds_data_t *value, Dataset *b)
 {
 	FileShare *a_file = value->data;
 	FileShare *b_file;
-	ds_data_t *b_key;
+	DatasetNode *node;
 
 	/*
 	 * The file doesn't exist in the new set at all, therefore it must've
 	 * been removed and we can safely return FALSE without any further
 	 * checks.
 	 */
-	if (!(b_key = dataset_lookup_ex (b, key)))
-		return FALSE;
+	if (!(node = dataset_lookup_node_ex (b, key)))
+		return NULL;
 
-	b_file = b_key->data;
+	b_file = node->value->data;
 
 	/*
 	 * If the mtime's of the files differ, giFT would've recalculated
 	 * new data and these cannot be considered the same any longer.
 	 */
-	return (SHARE_DATA(a_file)->mtime == SHARE_DATA(b_file)->mtime);
+	if (SHARE_DATA(a_file)->mtime != SHARE_DATA(b_file)->mtime)
+		return NULL;
+
+	return node;
 }
 
 static void emit_share_remove (ds_data_t *key, ds_data_t *value,
@@ -1075,7 +1026,14 @@ static void emit_share_remove_all (ds_data_t *key, ds_data_t *value,
                                    Dataset *latter)
 {
 	if (!eq_filesets (key, value, latter))
+	{
+#if 0
+		Share *t=value->data;
+		GIFT_TRACE (("removed %s", t->path));
+#endif
 		plugin_foreach (DS_FOREACH(emit_share_remove), value->data);
+		plugin_foreach (DS_FOREACH(emit_share_free), value->data);
+	}
 }
 
 static void emit_share_add (ds_data_t *key, ds_data_t *value, FileShare *file)
@@ -1094,7 +1052,14 @@ static int emit_share_add_all (ds_data_t *key, ds_data_t *value,
                                Dataset *former)
 {
 	if (!eq_filesets (key, value, former))
+	{
+#if 0
+		Share *t = value->data;
+		GIFT_TRACE (("added %s", t->path));
+#endif
+		plugin_foreach (DS_FOREACH(emit_share_new), value->data);
 		plugin_foreach (DS_FOREACH(emit_share_add), value->data);
+	}
 
 	return FALSE;
 }
@@ -1105,11 +1070,13 @@ static void handle_protocols (Dataset *former, Dataset *latter)
 {
 	int begin;
 
+#if 0
 	/*
 	 * Loop through all shares in the "new" set, emitting a share_new call
 	 * to each protocol for each file in this set.
 	 */
 	dataset_foreach (latter, DS_FOREACH(emit_share_new_all), NULL);
+#endif
 
 	/*
 	 * Now we must inform the protocol of all the modifications that occurred
@@ -1121,11 +1088,37 @@ static void handle_protocols (Dataset *former, Dataset *latter)
 	dataset_foreach (latter, DS_FOREACH(emit_share_add_all), former);
 	begin = FALSE; plugin_foreach (DS_FOREACH(emit_share_sync), &begin);
 
+#if 0
 	/*
 	 * We must now inform the protocol of all the structures which will soon
 	 * be deleted.
 	 */
 	dataset_foreach (former, DS_FOREACH(emit_share_free_all), NULL);
+#endif
+}
+
+/*****************************************************************************/
+
+static int flip_equal (ds_data_t *key, ds_data_t *value,
+                               Dataset *former)
+{
+	DatasetNode *node;
+
+	if ((node = eq_filesets (key, value, former)))
+	{
+		ds_data_t *temp = value->data;
+
+		value->data = node->value->data;
+		node->value->data = temp;
+	}
+
+	return FALSE;
+}
+
+/* no comment */
+static void twiddle_pointers (Dataset *former, Dataset *latter)
+{
+	dataset_foreach (former, DS_FOREACH(flip_equal), latter);
 }
 
 /*****************************************************************************/
@@ -1152,6 +1145,8 @@ Dataset *share_read_index (void)
 	share_build_index (&shares, &duplicates, sroot, &files, &size);
 	destroy_share_root (sroot);
 
+	twiddle_pointers (local_shares, shares);
+
 	handle_protocols (local_shares, shares);
 
 	/* clear any shares that existed before */
