@@ -1,5 +1,5 @@
 /*
- * $Id: upload.c,v 1.113 2004/01/15 06:28:56 hipnod Exp $
+ * $Id: upload.c,v 1.115 2004/03/22 22:19:46 mkern Exp $
  *
  * Copyright (C) 2001-2003 giFT project (gift.sourceforge.net)
  *
@@ -56,11 +56,18 @@ static Array    *uploads_queue = NULL; /* see queue_hit */
 /*
  * max_uploads:
  *
- *  0    no upload slots available
- * -1    unlimited
+ * >0    maximum number of uploads
+ *  0    unlimited
  */
-static int max_uploads = -1;
-static int disabled    = FALSE;
+static unsigned int max_uploads = 0;
+
+/*
+ * disabled:
+ *
+ * TRUE   Uploading disabled, shares hidden
+ * FALSE  Uploading enabled, shares shown
+ */
+static BOOL disabled    = FALSE;
 
 #define MAX_PERUSER_UPLOADS \
 	    config_get_int (gift_conf, "sharing/max_peruser_uploads=1")
@@ -612,7 +619,7 @@ static unsigned int queue_pos (unsigned int array_pos)
 	unsigned int available_slots = max_uploads;
 
 	/* should only be called if there's a queue to worry about */
-	assert (max_uploads >= 0);
+	assert (max_uploads > 0);
 
 	if (array_pos < available_slots)
 		return 0;
@@ -624,7 +631,7 @@ static int queue_ttl (void)
 {
 	unsigned int array_len = array_count (&uploads_queue);
 
-	assert (max_uploads >= 0);
+	assert (max_uploads > 0);
 
 	if (array_len < max_uploads)
 		return 0;
@@ -818,7 +825,7 @@ static int test_max_uploads (Protocol *p, const char *user, Share *share,
 			return UPLOAD_AUTH_MAX_PERUSER;
 	}
 
-	if (max_uploads == 0 || disabled)
+	if (disabled)
 		return UPLOAD_AUTH_HIDDEN;
 
 	/* add this request to the upload queue */
@@ -828,7 +835,7 @@ static int test_max_uploads (Protocol *p, const char *user, Share *share,
 	 * NOTE: If upload_availability() returns > 0 right now, we may still not
 	 * allow the upload if we are reserving an upload slot for somebody else.
 	 */
-	if (max_uploads < 0 || pos < max_uploads)
+	if (max_uploads == 0 || pos < max_uploads)
 		return UPLOAD_AUTH_ALLOW;
 
 	/* set the queued status information for the caller */
@@ -856,25 +863,66 @@ int upload_auth (Protocol *p, const char *user, Share *share,
 
 /*****************************************************************************/
 
+static void send_hide (ds_data_t *key, ds_data_t *value, unsigned long *avail)
+{
+	Protocol *p = value->data;
+
+	p->upload_avail (p, *avail);
+	p->share_hide (p);
+}
+
+static void send_show (ds_data_t *key, ds_data_t *value, unsigned long *avail)
+{
+	Protocol *p = value->data;
+
+	p->upload_avail (p, *avail);
+	p->share_show (p);
+}
+
 void upload_disable ()
 {
+	unsigned long avail;
+
+	/* save new state */
 	disabled = TRUE;
-	max_uploads = upload_status ();
+#if 0
+	config_set_int (gift_conf, "sharing/shares_hidden", disabled);
+#endif
+
+	/* hide shares */
+	avail = upload_availability ();
+	plugin_foreach (DS_FOREACH(send_hide), &avail);
 }
 
 void upload_enable ()
 {
+	unsigned long avail;
+
+	/* save new state */
 	disabled = FALSE;
-	max_uploads = upload_status ();
+#if 0
+	config_set_int (gift_conf, "sharing/shares_hidden", disabled);
+#endif
+
+	/* show shares */
+	avail = upload_availability ();
+	plugin_foreach (DS_FOREACH(send_show), &avail);
 }
 
 int upload_status ()
 {
-	if (disabled)
-		return 0;
-
 	/* sync with config */
-	max_uploads = config_get_int (gift_conf, "sharing/max_uploads=-1");
+#if 0
+	disabled    = config_get_int (gift_conf, "sharing/shares_hidden=0");
+#endif
+	max_uploads = config_get_int (gift_conf, "sharing/max_uploads=0");
+
+	/* we allow negative values in the config for backwards compatibility */
+	if(max_uploads < 0)
+		max_uploads = 0;
+
+	if (disabled)
+		return -1;
 
 	return max_uploads;
 }
@@ -886,12 +934,12 @@ unsigned long upload_availability (void)
 
 	status = upload_status ();
 
-	if (status == 0)
+	if (status < 0)
 		return av;
 
 	/* hardcode an availability of 9 (arbitrarily high number) in the event
 	 * that there is no limit imposed */
-	if (status < 0)
+	if (status == 0)
 		av = 9;
 	else
 		av = MAX (0, status - upload_length (NULL));
